"""
"""
from collections import OrderedDict

import numpy as np
from diffstar.fitting_helpers.fitting_kernels import _integrate_sfr
from dsps.constants import SFR_MIN
from dsps.experimental.diffburst import (
    _age_weights_from_params as _burst_age_weights_from_params,
)
from dsps.sed.stellar_age_weights import calc_age_weights_from_sfh_table
from dsps.utils import _jax_get_dt_array
from jax import jit as jjit
from jax import lax
from jax import numpy as jnp
from jax import vmap

from .disk_knots import _disk_knot_kern, _disk_knot_vmap

FBULGE_MIN = 0.05
FBULGE_MAX = 0.95

BOUNDING_K = 0.1

DEFAULT_FBULGE_EARLY = 0.75
DEFAULT_FBULGE_LATE = 0.15


DEFAULT_T10, DEFAULT_T90 = 2.0, 9.0
DEFAULT_FBULGE_PDICT = OrderedDict(fbulge_tcrit=8.0, fbulge_early=0.5, fbulge_late=0.1)
DEFAULT_FBULGE_PARAMS = np.array(list(DEFAULT_FBULGE_PDICT.values()))


_linterp_vmap = jjit(vmap(jnp.interp, in_axes=(0, None, 0)))

_A = (None, 0)
_burst_age_weights_from_params_vmap = jjit(
    vmap(_burst_age_weights_from_params, in_axes=_A)
)

_D = (None, 0, None, 0)
calc_age_weights_from_sfh_table_vmap = jjit(
    vmap(calc_age_weights_from_sfh_table, in_axes=_D)
)


@jjit
def _sigmoid(x, x0, k, ymin, ymax):
    height_diff = ymax - ymin
    return ymin + height_diff * lax.logistic(k * (x - x0))


@jjit
def _inverse_sigmoid(y, x0, k, ymin, ymax):
    lnarg = (ymax - ymin) / (y - ymin) - 1
    return x0 - lax.log(lnarg) / k


@jjit
def _tw_cuml_kern(x, m, h):
    """Triweight kernel version of an err function."""
    z = (x - m) / h
    val = (
        -5 * z**7 / 69984
        + 7 * z**5 / 2592
        - 35 * z**3 / 864
        + 35 * z / 96
        + 1 / 2
    )
    val = jnp.where(z < -3, 0, val)
    val = jnp.where(z > 3, 1, val)
    return val


@jjit
def _tw_sigmoid(x, x0, tw_h, ymin, ymax):
    height_diff = ymax - ymin
    body = _tw_cuml_kern(x, x0, tw_h)
    return ymin + height_diff * body


@jjit
def _bulge_fraction_kernel(t, thalf, frac_early, frac_late, dt):
    """typical values of 10.0, 0.7, 0.1
    frac_late < frac_late is needed to make bulges redder and so that
    bulge fractions increase with stellar mass
    """
    tw_h = dt / 6.0
    return _tw_sigmoid(t, thalf, tw_h, frac_early, frac_late)


@jjit
def _get_u_params_from_params(params, t10, t90):
    fbulge_tcrit, fbulge_early, fbulge_late = params

    t50 = (t10 + t90) / 2
    u_fbulge_tcrit = _inverse_sigmoid(fbulge_tcrit, t50, BOUNDING_K, t10, t90)

    x0 = (FBULGE_MIN + FBULGE_MAX) / 2
    u_fbulge_early = _inverse_sigmoid(
        fbulge_early, x0, BOUNDING_K, FBULGE_MIN, FBULGE_MAX
    )

    x0_late = (fbulge_early + FBULGE_MIN) / 2
    u_fbulge_late = _inverse_sigmoid(
        fbulge_late, x0_late, BOUNDING_K, fbulge_early, FBULGE_MIN
    )

    u_params = u_fbulge_tcrit, u_fbulge_early, u_fbulge_late
    return u_params


@jjit
def _get_params_from_u_params(u_params, t10, t90):
    u_fbulge_tcrit, u_fbulge_early, u_fbulge_late = u_params

    t50 = (t10 + t90) / 2
    fbulge_tcrit = _sigmoid(u_fbulge_tcrit, t50, BOUNDING_K, t10, t90)

    x0 = (FBULGE_MIN + FBULGE_MAX) / 2
    fbulge_early = _sigmoid(u_fbulge_early, x0, BOUNDING_K, FBULGE_MIN, FBULGE_MAX)

    x0_late = (fbulge_early + FBULGE_MIN) / 2
    fbulge_late = _sigmoid(u_fbulge_late, x0_late, BOUNDING_K, fbulge_early, FBULGE_MIN)

    params = fbulge_tcrit, fbulge_early, fbulge_late
    return params


@jjit
def _bulge_fraction_vs_tform_u_params(t, t10, t90, u_params):
    params = _get_params_from_u_params(u_params, t10, t90)
    fbulge_tcrit, fbulge_early, fbulge_late = params
    dt = t90 - t10
    return _bulge_fraction_kernel(t, fbulge_tcrit, fbulge_early, fbulge_late, dt)


@jjit
def calc_tform_kern(abscissa, xarr, tform_frac):
    fracarr = xarr / xarr[-1]
    return jnp.interp(tform_frac, fracarr, abscissa)


_calc_tform_pop_kern = jjit(vmap(calc_tform_kern, in_axes=[None, 0, None]))


@jjit
def calc_tform_pop(tarr, smh_pop, tform_frac):
    """Calculate the formation time of a population

    Parameters
    ----------
    tarr : ndarray, shape(nt, )

    smh_pop : ndarray, shape(npop, nt)

    tform_frac : float
        Fraction used in the formation time definition
        tform_frac=0.5 corresponds to the half-mass time, for example

    """
    return _calc_tform_pop_kern(tarr, smh_pop, tform_frac)


@jjit
def _bulge_fraction_vs_tform(t, t10, t90, params):
    fbulge_tcrit, fbulge_early, fbulge_late = params
    dt = t90 - t10
    fbulge = _bulge_fraction_kernel(t, fbulge_tcrit, fbulge_early, fbulge_late, dt)
    return fbulge


@jjit
def _bulge_sfh(tarr, sfh, fbulge_params):
    dtarr = _jax_get_dt_array(tarr)
    sfh = jnp.where(sfh < SFR_MIN, SFR_MIN, sfh)
    smh = _integrate_sfr(sfh, dtarr)
    fracmh = smh / smh[-1]
    t10 = jnp.interp(0.1, fracmh, tarr)
    t90 = jnp.interp(0.9, fracmh, tarr)
    eff_bulge = _bulge_fraction_vs_tform(tarr, t10, t90, fbulge_params)
    sfh_bulge = eff_bulge * sfh
    smh_bulge = _integrate_sfr(sfh_bulge, dtarr)
    bth = smh_bulge / smh
    return smh, eff_bulge, sfh_bulge, smh_bulge, bth


_B = (None, 0, 0)
_bulge_sfh_vmap = jjit(vmap(_bulge_sfh, in_axes=_B))


def decompose_sfhpop_into_bulge_disk_knots(
    gal_fbulge_params,
    gal_fknot,
    gal_t_obs,
    gal_t_table,
    gal_sfh,
    gal_fburst,
    gal_burstshape_params,
    ssp_lg_age_gyr,
):
    """Decompose the SFH of a Diffsky galaxy into three components:
    bulges, diffuse disk, and star-forming knots in the disks

    Parameters
    ----------
    gal_fbulge_params : ndarray, shape (n_gals, 3)
        Bulge efficiency parameters (tcrit, fbulge_early, fbulge_late) for each galaxy

    gal_fknot : ndarray, shape (n_gals, )
        Fraction of the disk mass in bursty star-forming knots for each galaxy

    gal_t_obs : ndarray, shape (n_gals, )
        Age of the universe in Gyr at the redshift of each galaxy

    gal_t_table : ndarray, shape (n_t, )
        Grid in cosmic time t in Gyr at which SFH of the galaxy population is tabulated
        gal_t_table should increase monotonically and it should span the
        full range of gal_t_obs, including some padding of a few million years

    gal_sfh : ndarray, shape (n_gals, n_t)
        Grid in SFR in Msun/yr for each galaxy tabulated at the input gal_t_table

    gal_fburst : ndarray, shape (n_gals, )
        Fraction of stellar mass in the burst population of each galaxy

    gal_burstshape_params : ndarray, shape (n_gals, 2)
        Parameters controlling P(τ) for burst population in each galaxy
        lgyr_peak = gal_burstshape_params[:, 0]
        lgyr_max = gal_burstshape_params[:, 1]

    ssp_lg_age_gyr : ndarray, shape (n_age, )
        Grid in age τ at which the SSPs are computed, stored as log10(τ/Gyr)

    Returns
    -------
    mbulge : ndarray, shape (n_gals, )
        Total stellar mass in Msun formed in the bulge at time gal_t_obs

    mdd : ndarray, shape (n_gals, )
        Total stellar mass in Msun formed in the diffuse disk at time gal_t_obs

    mknot : ndarray, shape (n_gals, )
        Total stellar mass in Msun formed in star-forming knots at time gal_t_obs

    mburst : ndarray, shape (n_gals, )
        Total stellar mass in Msun in the burst population at time gal_t_obs

    bulge_age_weights : ndarray, shape (n_gals, n_age)
        Probability distribution P(τ_age) for bulge of each galaxy

    dd_age_weights : ndarray, shape (n_gals, n_age)
        Probability distribution P(τ_age) for diffuse disk of each galaxy

    knot_age_weights : ndarray, shape (n_gals, n_age)
        Probability distribution P(τ_age) for star-forming knots of each galaxy

    bulge_sfh : ndarray, shape (n_gals, n_t)
        Grid in SFR in Msun/yr for each galaxy bulge tabulated at the input gal_t_table

    gal_frac_bulge_t_obs : ndarray, shape (n_gals, )
        Bulge/total mass ratio at gal_t_obs for every galaxy

    """
    ssp_lg_age_yr = ssp_lg_age_gyr + 9.0
    gal_burst_age_weights = _burst_age_weights_from_params_vmap(
        ssp_lg_age_yr, gal_burstshape_params
    )
    return _decompose_sfhpop_into_bulge_disk_knots(
        gal_fbulge_params,
        gal_fknot,
        gal_t_obs,
        gal_t_table,
        gal_sfh,
        gal_fburst,
        gal_burst_age_weights,
        ssp_lg_age_gyr,
    )


@jjit
def _decompose_sfh_singlegal_into_bulge_disk_knots(
    fbulge_params,
    fknot,
    t_obs,
    t_table,
    sfh_table,
    fburst,
    age_weights_burst,
    ssp_lg_age_gyr,
):
    _res = _bulge_sfh(t_table, sfh_table, fbulge_params)
    smh, eff_bulge, bulge_sfh, smh_bulge, bulge_to_total_history = _res

    bulge_sfh = jnp.where(bulge_sfh < SFR_MIN, SFR_MIN, bulge_sfh)
    frac_bulge_t_obs = jnp.interp(t_obs, t_table, bulge_to_total_history)

    bulge_age_weights = calc_age_weights_from_sfh_table(
        t_table, bulge_sfh, ssp_lg_age_gyr, t_obs
    )
    disk_sfh = sfh_table - bulge_sfh
    disk_sfh = jnp.where(disk_sfh < SFR_MIN, SFR_MIN, disk_sfh)

    args = (
        t_table,
        t_obs,
        sfh_table,
        disk_sfh,
        fburst,
        fknot,
        age_weights_burst,
        ssp_lg_age_gyr,
    )
    _knot_info = _disk_knot_kern(*args)
    mstar_tot, mburst, mdd, mknot, dd_age_weights, knot_age_weights = _knot_info

    mbulge = frac_bulge_t_obs * mstar_tot
    masses = mbulge, mdd, mknot, mburst
    age_weights = bulge_age_weights, dd_age_weights, knot_age_weights
    ret = (*masses, *age_weights, bulge_sfh, frac_bulge_t_obs)
    return ret


@jjit
def _decompose_sfhpop_into_bulge_disk_knots(
    gal_fbulge_params,
    gal_fknot,
    gal_t_obs,
    gal_t_table,
    gal_sfh,
    gal_fburst,
    age_weights_burst,
    ssp_lg_age_gyr,
):
    _res = _bulge_sfh_vmap(gal_t_table, gal_sfh, gal_fbulge_params)
    smh, eff_bulge, bulge_sfh, smh_bulge, bulge_to_total_history = _res

    bulge_sfh = jnp.where(bulge_sfh < SFR_MIN, SFR_MIN, bulge_sfh)
    gal_frac_bulge_t_obs = _linterp_vmap(gal_t_obs, gal_t_table, bulge_to_total_history)

    bulge_age_weights = calc_age_weights_from_sfh_table_vmap(
        gal_t_table, bulge_sfh, ssp_lg_age_gyr, gal_t_obs
    )

    disk_sfh = gal_sfh - bulge_sfh
    disk_sfh = jnp.where(disk_sfh < SFR_MIN, SFR_MIN, disk_sfh)

    args = (
        gal_t_table,
        gal_t_obs,
        gal_sfh,
        disk_sfh,
        gal_fburst,
        gal_fknot,
        age_weights_burst,
        ssp_lg_age_gyr,
    )
    _knot_info = _disk_knot_vmap(*args)
    mstar_tot, mburst, mdd, mknot, dd_age_weights, knot_age_weights = _knot_info

    mbulge = gal_frac_bulge_t_obs * mstar_tot
    masses = mbulge, mdd, mknot, mburst
    age_weights = bulge_age_weights, dd_age_weights, knot_age_weights
    ret = (*masses, *age_weights, bulge_sfh, gal_frac_bulge_t_obs)
    return ret


DEFAULT_FBULGE_U_PARAMS = _get_u_params_from_params(
    DEFAULT_FBULGE_PARAMS, DEFAULT_T10, DEFAULT_T90
)
_A = (0, 0, 0)
_get_params_from_u_params_vmap = jjit(vmap(_get_params_from_u_params, in_axes=_A))
